/*
 *     Module Name       : JDMSocket.cpp
 *
 *     Author/Location   : J.R.F. De Maeijer, Amstelveen
 *
 *     Copyright         : 2011, J.R.F. De Maeijer, Amstelveen
 *
 * ----------------------------------------------------------------------------
 * Copyright (c) 2011, J.R.F. De Maeijer
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ----------------------------------------------------------------------------
 *                            MODIFICATION HISTORY
 * ----------------------------------------------------------------------------
 * DATE        REASON                                                    AUTHOR
 * ----------------------------------------------------------------------------
 * 16-Jan-2011 Initial Release                                           J.D.M.
 * 05-Apr-2011 Renamed to JDMSocket and implemented GetStatusText()      J.D.M.
 * 15-Jul-2013 - Added functions GetMaximumMessageSizeRead() and
 *               GetMaximumMessageSizeWrite()
 *             - Renamed JDM_SOCKET to JDMSOCKET
 *             - Changed __cdecl into __stdcall at the interface         J.D.M.
 * 19-Jul-2013 - Adapted read and write functions
 *             - Added resource file V1.0.1.                             J.D.M.
 * 13-Dec-2013 - Added GetInitFileDescriptor() and moved functions
 *               GetMaximumMessageSizeRead/Write() to interface.
 *             - Added some defensive programming regarding checking of
 *               socket file descriptors.                                J.D.M.
 * 26-Mar-2014 - Added JDMSOCKET_ERROR_COULD_NOT_SET_SOCKET_OPTION
 *             - Moved code in JDMSocket::DeInit() for correct behaviour
 *             - Added SO_REUSEADDR to JDMSocket::Listen()               J.D.M.
 * ----------------------------------------------------------------------------
 */

#include <stdlib.h>

#include "JDMSocket.h"

#define JDMSOCKET_TYPE_SERVER  0 /* Socket */
#define JDMSOCKET_TYPE_CLIENT  1 /* Plug */

#define JDMSOCKET_BLOCKING_MODE_ON   0
#define JDMSOCKET_BLOCKING_MODE_OFF  1

/*---------------------------------------------------------------------------*/
/* JDMSocket::JDMSocket                                                      */
/*---------------------------------------------------------------------------*/
JDMSocket::JDMSocket()
          :Port(-1)
          ,ServerFD(0)
          ,LastStatus(0)
          ,LastLengthWritten(0)
          ,LastLengthRead(0)
{
} /* JDMSocket::JDMSocket */

/*---------------------------------------------------------------------------*/
/* JDMSocket::~JDMSocket                                                     */
/*---------------------------------------------------------------------------*/
JDMSocket::~JDMSocket()
{
  StatusText="";
  ErrorText="";
  DeInit();
} /* JDMSocket::~JDMSocket */

/*---------------------------------------------------------------------------*/
/* JDMSocket::Init                                                           */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::Init(long Port, bool Blocking)
{
  return Init(Port, Blocking, 0);
} /* JDMSocket::Init */

/*---------------------------------------------------------------------------*/
/* JDMSocket::Init                                                           */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::Init(long Port, bool Blocking, const char* HostName)
{
  WSASetLastError(0);
  SaveStatus(0);
  LastLengthRead=0;
  LastLengthWritten=0;

  if(HostName!=0) {
    Type=JDMSOCKET_TYPE_CLIENT;
  }
  else {
    Type=JDMSOCKET_TYPE_SERVER;
  }

  this->Port = Port;
  this->BlockingMode = Blocking==true?JDMSOCKET_BLOCKING_MODE_ON:JDMSOCKET_BLOCKING_MODE_OFF;

  memset(&ServerAddress, 0, sizeof(ServerAddress));
  ServerAddress.sin_family = AF_INET;
  ServerAddress.sin_port   = htons((unsigned short)Port);
  if(Type==JDMSOCKET_TYPE_CLIENT) {
    struct hostent* Host;
    Host=gethostbyname(HostName);
    if(Host==0) {
      SaveStatus(WSAGetLastError());
      return JDMSOCKET_ERROR_GETHOSTBYNAME;
    }
    memcpy((void*)&ServerAddress.sin_addr, (void*)Host->h_addr, Host->h_length);
    return Connect();
  }
  else {
    return Listen();
  }
} /* JDMSocket::Init */

/*---------------------------------------------------------------------------*/
/* JDMSocket::DeInit                                                         */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::DeInit()
{
  if(ServerFD>0 && closesocket(ServerFD)==SOCKET_ERROR) {
    if(WSAGetLastError()==0) {SaveStatus(WSAGetLastError());}
    return JDMSOCKET_ERROR_COULD_NOT_CLOSE;
  }

  ServerFD=0;
  Port=-1;
  memset(&ServerAddress, 0, sizeof(ServerAddress));
  BlockingMode=JDMSOCKET_BLOCKING_MODE_ON;
  LastLengthRead=0;
  LastLengthWritten=0;

  if(WSAGetLastError()==0) {SaveStatus(WSAGetLastError());}
  return JDMSOCKET_NO_ERROR;
} /* JDMSocket::DeInit */

/*---------------------------------------------------------------------------*/
/* JDMSocket::Read                                                           */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::Read(char* RawData, long MaxDataLength, bool WhenBlockingWaitForMaxDataLength)
{
  WSASetLastError(0);
  SaveStatus(0);
  LastLengthRead=0;

  if(Type==JDMSOCKET_TYPE_CLIENT && ServerFD>0) {
    return Read(ServerFD, RawData, MaxDataLength, WhenBlockingWaitForMaxDataLength);
  }

  return JDMSOCKET_ERROR_NO_FILEDESCRIPTOR;
} /* JDMSocket::Read */

/*---------------------------------------------------------------------------*/
/* JDMSocket::Read                                                           */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::Read(long FileDescriptor, char* RawData, long MaxDataLength, bool WhenBlockingWaitForMaxDataLength)
{
  return ReadFromSocket(FileDescriptor, RawData, MaxDataLength, WhenBlockingWaitForMaxDataLength);
} /* JDMSocket::Read */

/*---------------------------------------------------------------------------*/
/* JDMSocket::ReadFromSocket                                                 */
/*---------------------------------------------------------------------------*/
long JDMSocket::ReadFromSocket(long FileDescriptor, char* RawData, long MaxDataLength, bool WhenBlockingWaitForMaxDataLength)
{
  long RetVal;
  long LengthRead=0;
  long LeftToRead=MaxDataLength;

  WSASetLastError(0);
  SaveStatus(0);
  LastLengthRead=0;

  if(FileDescriptor<=0) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_NO_FILEDESCRIPTOR;
  }
  if(RawData==0||MaxDataLength<=0) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_NO_DATA_TO_READ;
  }
  if(MaxDataLength>GetMaximumMessageSizeRead(FileDescriptor)) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_LENGTH_DATA_EXCEEDED;
  }
  if(WhenBlockingWaitForMaxDataLength==true) {
    while(LeftToRead > 0) {
      RetVal=recv(FileDescriptor, (char*)RawData+LengthRead, LeftToRead, 0);
      if(RetVal==SOCKET_ERROR||RetVal==0) {
        break;
      }
      LengthRead+=RetVal;
      LeftToRead-=RetVal;
    }
  }
  else {
    RetVal=recv(FileDescriptor, (char*)RawData+LengthRead, LeftToRead, 0);
    if(RetVal!=SOCKET_ERROR && RetVal!=0) {
      LengthRead+=RetVal;
      LeftToRead-=RetVal;
    }
  }
  LastLengthRead=LengthRead;

  SaveStatus(WSAGetLastError());
  if(RetVal==SOCKET_ERROR) {
    if(LastStatus==WSAEWOULDBLOCK) {
      WSASetLastError(0);
      SaveStatus(0);
      return LengthRead; /* returns number of bytes read */
    }
    return JDMSOCKET_ERROR_DURING_READ;
  }
  if(RetVal==0) {
    /* Connection is closed */
    return JDMSOCKET_ERROR_CONNECTION_CLOSED;
  }

  if(LengthRead==0 || (WhenBlockingWaitForMaxDataLength==true && LengthRead!=MaxDataLength)) {
    return JDMSOCKET_ERROR_DURING_READ;
  }
  else {
    return LengthRead; /* returns number of bytes read */
  }
} /* JDMSocket::ReadFromSocket */

/*---------------------------------------------------------------------------*/
/* JDMSocket::Write                                                          */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::Write(const char* RawData, long DataLength)
{
  WSASetLastError(0);
  SaveStatus(0);
  LastLengthWritten=0;

  if(Type==JDMSOCKET_TYPE_CLIENT && ServerFD>0) {
    return WriteToSocket(ServerFD, RawData, DataLength);
  }

  return JDMSOCKET_ERROR_NO_FILEDESCRIPTOR;
} /* JDMSocket::Write */

/*---------------------------------------------------------------------------*/
/* JDMSocket::Write                                                          */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::Write(long FileDescriptor, const char* RawData, long DataLength)
{
  return WriteToSocket(FileDescriptor, RawData, DataLength);
} /* JDMSocket::Write */

/*---------------------------------------------------------------------------*/
/* JDMSocket::WriteToSocket                                                  */
/*---------------------------------------------------------------------------*/
long JDMSocket::WriteToSocket(long FileDescriptor, const char* RawData, long DataLength)
{
  long RetVal;
  long LengthWritten=0;
  long LeftToWrite=DataLength;

  WSASetLastError(0);
  SaveStatus(0);
  LastLengthWritten=0;

  if(FileDescriptor<=0) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_NO_FILEDESCRIPTOR;
  }
  if(RawData==0||DataLength<=0) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_NO_DATA_TO_WRITE;
  }
  if(DataLength>GetMaximumMessageSizeWrite(FileDescriptor)) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_LENGTH_DATA_EXCEEDED;
  }
  while(LengthWritten < LeftToWrite) {
    RetVal=send(FileDescriptor, RawData+LengthWritten, LeftToWrite, 0);
    if(RetVal==SOCKET_ERROR) {
      break;
    }
    LengthWritten+=RetVal;
    LeftToWrite-=RetVal;
  }
  LastLengthWritten=LengthWritten;

  SaveStatus(WSAGetLastError());
  if(RetVal==SOCKET_ERROR) {
    if(LastStatus==WSAEWOULDBLOCK) { /* Not enough space available on the socket */
      WSASetLastError(0);
      SaveStatus(0);
      return JDMSOCKET_WAIT_AND_SEND_REMAINING_AGAIN;
    }
  }

  if(LengthWritten==0 || LengthWritten!=DataLength) {
    return JDMSOCKET_ERROR_DURING_SEND;
  }
  else {
    return JDMSOCKET_NO_ERROR;
  }
} /* JDMSocket::WriteToSocket */

/*---------------------------------------------------------------------------*/
/* JDMSocket::AcceptConnection                                               */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::AcceptConnection()
{
  WSASetLastError(0);
  SaveStatus(0);

  if(Type==JDMSOCKET_TYPE_CLIENT) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_COULD_NOT_ACCEPT;
  }
  if(ServerFD<=0) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_NO_FILEDESCRIPTOR;
  }

  long ClientFD=0;
  struct sockaddr_in ClientAddress;
  int AddressLength=sizeof(struct sockaddr_in);
  memset(&ClientAddress, 0, sizeof(ClientAddress));

  ClientFD=accept(ServerFD, (struct sockaddr*)&ClientAddress, &AddressLength);
  if(ClientFD==INVALID_SOCKET) {
    ClientFD=0;
    SaveStatus(WSAGetLastError());
    if(LastStatus==WSAEINTR||LastStatus==WSATRY_AGAIN) {
      return JDMSOCKET_ERROR_NO_DATA;
    }
    return JDMSOCKET_ERROR_COULD_NOT_ACCEPT;
  }

  SaveStatus(WSAGetLastError());
  return ClientFD;
} /* JDMSocket::AcceptConnection */

/*---------------------------------------------------------------------------*/
/* JDMSocket::CloseConnection                                                */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::CloseConnection(long& FileDescriptor)
{
  WSASetLastError(0);
  SaveStatus(0);

  if(FileDescriptor<=0) {
    FileDescriptor=0;
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_NO_ERROR;
  }
  if(FileDescriptor==ServerFD) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_USE_SOCKET_DEINIT;
  }
  if(closesocket(FileDescriptor)==SOCKET_ERROR) {
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_COULD_NOT_CLOSE;
  }

  FileDescriptor=0;

  SaveStatus(WSAGetLastError());
  return JDMSOCKET_NO_ERROR;
} /* JDMSocket::CloseConnection */

/*---------------------------------------------------------------------------*/
/* JDMSocket::GetMaximumMessageSizeRead                                      */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::GetMaximumMessageSizeRead(long FileDescriptor)
{
  int Value=0;
  int ValueSize=sizeof(Value);

  WSASetLastError(0);
  SaveStatus(0);

  if(FileDescriptor<=0) {
    SaveStatus(WSAGetLastError());
    return 0;
  }
  if(getsockopt(FileDescriptor, SOL_SOCKET, SO_RCVBUF, (char*)&Value, &ValueSize)==SOCKET_ERROR) {
    SaveStatus(WSAGetLastError());
    return 0;
  }

  return Value;
} /* JDMSocket::GetMaximumMessageSizeRead */

/*---------------------------------------------------------------------------*/
/* JDMSocket::GetMaximumMessageSizeWrite                                     */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::GetMaximumMessageSizeWrite(long FileDescriptor)
{
  int Value=0;
  int ValueSize=sizeof(Value);

  WSASetLastError(0);
  SaveStatus(0);

  if(FileDescriptor<=0) {
    SaveStatus(WSAGetLastError());
    return 0;
  }
  if(getsockopt(FileDescriptor, SOL_SOCKET, SO_SNDBUF, (char*)&Value, &ValueSize)==SOCKET_ERROR) {
    SaveStatus(WSAGetLastError());
    return 0;
  }

  return Value;
} /* JDMSocket::GetMaximumMessageSizeWrite */

/*---------------------------------------------------------------------------*/
/* JDMSocket::GetLastStatus                                                  */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::GetLastStatus() {
  return LastStatus;
} /* JDMSocket::GetLastStatus */

/*---------------------------------------------------------------------------*/
/* JDMSocket::GetLastLengthWritten                                           */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::GetLastLengthWritten() {
  return LastLengthWritten;
} /* JDMSocket::GetLastLengthWritten */

/*---------------------------------------------------------------------------*/
/* JDMSocket::GetLastLengthRead                                              */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::GetLastLengthRead() {
  return LastLengthRead;
} /* JDMSocket::GetLastLengthRead */

/*---------------------------------------------------------------------------*/
/* JDMSocket::GetStatusText                                                  */
/*---------------------------------------------------------------------------*/
const char* JDMSOCKET_CALL_CONV JDMSocket::GetStatusText(long StatusValue) {
  string TempText;
  char   FormatString[128];

  switch(StatusValue) { /* http://msdn.microsoft.com/en-us/library/ms740668(v=vs.85).aspx */
    case 0:                                 TempText="NO ERROR"; break;
    case ERROR_FILE_NOT_FOUND:              TempText="ERROR_FILE_NOT_FOUND"; break;
    case ERROR_PATH_NOT_FOUND:              TempText="ERROR_PATH_NOT_FOUND"; break;
    case ERROR_ACCESS_DENIED:               TempText="ERROR_ACCESS_DENIED"; break;
    case ERROR_INVALID_HANDLE:              TempText="ERROR_INVALID_HANDLE"; break;
    case ERROR_NOT_ENOUGH_MEMORY:           TempText="ERROR_NOT_ENOUGH_MEMORY"; break;
    case ERROR_SHARING_VIOLATION:           TempText="ERROR_SHARING_VIOLATION"; break;
    case ERROR_INVALID_PARAMETER:           TempText="ERROR_INVALID_PARAMETER"; break;
    case ERROR_OPERATION_ABORTED:           TempText="ERROR_OPERATION_ABORTED"; break;
    case ERROR_IO_INCOMPLETE:               TempText="ERROR_IO_INCOMPLETE"; break;
    case ERROR_IO_PENDING:                  TempText="ERROR_IO_PENDING"; break;
    case WSAEINTR:                          TempText="WSAEINTR"; break;
    case WSAEBADF:                          TempText="WSAEBADF"; break;
    case WSAEACCES:                         TempText="WSAEACCES"; break;
    case WSAEFAULT:                         TempText="WSAEFAULT"; break;
    case WSAEINVAL:                         TempText="WSAEINVAL"; break;
    case WSAEMFILE:                         TempText="WSAEMFILE"; break;
    case WSAEWOULDBLOCK:                    TempText="WSAEWOULDBLOCK"; break;
    case WSAEINPROGRESS:                    TempText="WSAEINPROGRESS"; break;
    case WSAEALREADY:                       TempText="WSAEALREADY"; break;
    case WSAENOTSOCK:                       TempText="WSAENOTSOCK"; break;
    case WSAEDESTADDRREQ:                   TempText="WSAEDESTADDRREQ"; break;
    case WSAEMSGSIZE:                       TempText="WSAEMSGSIZE"; break;
    case WSAEPROTOTYPE:                     TempText="WSAEPROTOTYPE"; break;
    case WSAENOPROTOOPT:                    TempText="WSAENOPROTOOPT"; break;
    case WSAEPROTONOSUPPORT:                TempText="WSAEPROTONOSUPPORT"; break;
    case WSAESOCKTNOSUPPORT:                TempText="WSAESOCKTNOSUPPORT"; break;
    case WSAEOPNOTSUPP:                     TempText="WSAEOPNOTSUPP"; break;
    case WSAEPFNOSUPPORT:                   TempText="WSAEPFNOSUPPORT"; break;
    case WSAEAFNOSUPPORT:                   TempText="WSAEAFNOSUPPORT"; break;
    case WSAEADDRINUSE:                     TempText="WSAEADDRINUSE"; break;
    case WSAEADDRNOTAVAIL:                  TempText="WSAEADDRNOTAVAIL"; break;
    case WSAENETDOWN:                       TempText="WSAENETDOWN"; break;
    case WSAENETUNREACH:                    TempText="WSAENETUNREACH"; break;
    case WSAENETRESET:                      TempText="WSAENETRESET"; break;
    case WSAECONNABORTED:                   TempText="WSAECONNABORTED"; break;
    case WSAECONNRESET:                     TempText="WSAECONNRESET"; break;
    case WSAENOBUFS:                        TempText="WSAENOBUFS"; break;
    case WSAEISCONN:                        TempText="WSAEISCONN"; break;
    case WSAENOTCONN:                       TempText="WSAENOTCONN"; break;
    case WSAESHUTDOWN:                      TempText="WSAESHUTDOWN"; break;
    case WSAETOOMANYREFS:                   TempText="WSAETOOMANYREFS"; break;
    case WSAETIMEDOUT:                      TempText="WSAETIMEDOUT"; break;
    case WSAECONNREFUSED:                   TempText="WSAECONNREFUSED"; break;
    case WSAELOOP:                          TempText="WSAELOOP"; break;
    case WSAENAMETOOLONG:                   TempText="WSAENAMETOOLONG"; break;
    case WSAEHOSTDOWN:                      TempText="WSAEHOSTDOWN"; break;
    case WSAEHOSTUNREACH:                   TempText="WSAEHOSTUNREACH"; break;
    case WSAENOTEMPTY:                      TempText="WSAENOTEMPTY"; break;
    case WSAEPROCLIM:                       TempText="WSAEPROCLIM"; break;
    case WSAEUSERS:                         TempText="WSAEUSERS"; break;
    case WSAEDQUOT:                         TempText="WSAEDQUOT"; break;
    case WSAESTALE:                         TempText="WSAESTALE"; break;
    case WSAEREMOTE:                        TempText="WSAEREMOTE"; break;
    case WSASYSNOTREADY:                    TempText="WSASYSNOTREADY"; break;
    case WSAVERNOTSUPPORTED:                TempText="WSAVERNOTSUPPORTED"; break;
    case WSANOTINITIALISED:                 TempText="WSANOTINITIALISED"; break;
    case WSAEDISCON:                        TempText="WSAEDISCON"; break;
    case WSAENOMORE:                        TempText="WSAENOMORE"; break;
    case WSAECANCELLED:                     TempText="WSAECANCELLED"; break;
    case WSAEINVALIDPROCTABLE:              TempText="WSAEINVALIDPROCTABLE"; break;
    case WSAEINVALIDPROVIDER:               TempText="WSAEINVALIDPROVIDER"; break;
    case WSAEPROVIDERFAILEDINIT:            TempText="WSAEPROVIDERFAILEDINIT"; break;
    case WSASYSCALLFAILURE:                 TempText="WSASYSCALLFAILURE"; break;
    case WSASERVICE_NOT_FOUND:              TempText="WSASERVICE_NOT_FOUND"; break;
    case WSATYPE_NOT_FOUND:                 TempText="WSATYPE_NOT_FOUND"; break;
    case WSA_E_NO_MORE:                     TempText="WSA_E_NO_MORE"; break;
    case WSA_E_CANCELLED:                   TempText="WSA_E_CANCELLED"; break;
    case WSAEREFUSED:                       TempText="WSAEREFUSED"; break;
    case WSAHOST_NOT_FOUND:                 TempText="WSAHOST_NOT_FOUND"; break;
    case WSATRY_AGAIN:                      TempText="WSATRY_AGAIN"; break;
    case WSANO_RECOVERY:                    TempText="WSANO_RECOVERY"; break;
    case WSANO_DATA:                        TempText="WSANO_DATA"; break;
    case WSA_QOS_RECEIVERS:                 TempText="WSA_QOS_RECEIVERS"; break;
    case WSA_QOS_SENDERS:                   TempText="WSA_QOS_SENDERS"; break;
    case WSA_QOS_NO_SENDERS:                TempText="WSA_QOS_NO_SENDERS"; break;
    case WSA_QOS_NO_RECEIVERS:              TempText="WSA_QOS_NO_RECEIVERS"; break;
    case WSA_QOS_REQUEST_CONFIRMED:         TempText="WSA_QOS_REQUEST_CONFIRMED"; break;
    case WSA_QOS_ADMISSION_FAILURE:         TempText="WSA_QOS_ADMISSION_FAILURE"; break;
    case WSA_QOS_POLICY_FAILURE:            TempText="WSA_QOS_POLICY_FAILURE"; break;
    case WSA_QOS_BAD_STYLE:                 TempText="WSA_QOS_BAD_STYLE"; break;
    case WSA_QOS_BAD_OBJECT:                TempText="WSA_QOS_BAD_OBJECT"; break;
    case WSA_QOS_TRAFFIC_CTRL_ERROR:        TempText="WSA_QOS_TRAFFIC_CTRL_ERROR"; break;
    case WSA_QOS_GENERIC_ERROR:             TempText="WSA_QOS_GENERIC_ERROR"; break;
    case WSA_QOS_ESERVICETYPE:              TempText="WSA_QOS_ESERVICETYPE"; break;
    case WSA_QOS_EFLOWSPEC:                 TempText="WSA_QOS_EFLOWSPEC"; break;
    case WSA_QOS_EPROVSPECBUF:              TempText="WSA_QOS_EPROVSPECBUF"; break;
    case WSA_QOS_EFILTERSTYLE:              TempText="WSA_QOS_EFILTERSTYLE"; break;
    case WSA_QOS_EFILTERTYPE:               TempText="WSA_QOS_EFILTERTYPE"; break;
    case WSA_QOS_EFILTERCOUNT:              TempText="WSA_QOS_EFILTERCOUNT"; break;
    case WSA_QOS_EOBJLENGTH:                TempText="WSA_QOS_EOBJLENGTH"; break;
    case WSA_QOS_EFLOWCOUNT:                TempText="WSA_QOS_EFLOWCOUNT"; break;
    case WSA_QOS_EUNKOWNPSOBJ:              TempText="WSA_QOS_EUNKOWNPSOBJ"; break;
    case WSA_QOS_EPOLICYOBJ:                TempText="WSA_QOS_EPOLICYOBJ"; break;
    case WSA_QOS_EFLOWDESC:                 TempText="WSA_QOS_EFLOWDESC"; break;
    case WSA_QOS_EPSFLOWSPEC:               TempText="WSA_QOS_EPSFLOWSPEC"; break;
    case WSA_QOS_EPSFILTERSPEC:             TempText="WSA_QOS_EPSFILTERSPEC"; break;
    case WSA_QOS_ESDMODEOBJ:                TempText="WSA_QOS_ESDMODEOBJ"; break;
    case WSA_QOS_ESHAPERATEOBJ:             TempText="WSA_QOS_ESHAPERATEOBJ"; break;
    case WSA_QOS_RESERVED_PETYPE:           TempText="WSA_QOS_RESERVED_PETYPE"; break;
    default:                                TempText="UNKNOWN_STATUS"; break;
  }
  /* Used because of boundschecker warnings when wostringstream is used: */
  sprintf(FormatString, "(%+04.3ld) %.100s", StatusValue, TempText.c_str());
  StatusText=FormatString;

  return StatusText.c_str();
} /* JDMSocket::GetStatusText */

/*---------------------------------------------------------------------------*/
/* JDMSocket::GetErrorText                                                   */
/*---------------------------------------------------------------------------*/
const char* JDMSOCKET_CALL_CONV JDMSocket::GetErrorText(long ErrorValue) {
  string TempText;
  char   FormatString[128];

  switch(ErrorValue) {
    case JDMSOCKET_NO_ERROR:                          TempText="JDMSOCKET_NO_ERROR"; break;
    case JDMSOCKET_WAIT_AND_SEND_REMAINING_AGAIN:     TempText="JDMSOCKET_WAIT_AND_SEND_REMAINING_AGAIN"; break;
    case JDMSOCKET_WAIT_AND_READ_REMAINING_AGAIN:     TempText="JDMSOCKET_WAIT_AND_READ_REMAINING_AGAIN"; break;
    case JDMSOCKET_ERROR_WSASYSNOTREADY:              TempText="JDMSOCKET_ERROR_WSASYSNOTREADY"; break;
    case JDMSOCKET_ERROR_WSAVERNOTSUPPORTED:          TempText="JDMSOCKET_ERROR_WSAVERNOTSUPPORTED"; break;
    case JDMSOCKET_ERROR_WSAEINPROGRESS:              TempText="JDMSOCKET_ERROR_WSAEINPROGRESS"; break;
    case JDMSOCKET_ERROR_WSAEPROCLIM:                 TempText="JDMSOCKET_ERROR_WSAEPROCLIM"; break;
    case JDMSOCKET_ERROR_WSAEFAULT:                   TempText="JDMSOCKET_ERROR_WSAEFAULT"; break;
    case JDMSOCKET_ERROR_WSANOTINITIALISED:           TempText="JDMSOCKET_ERROR_WSANOTINITIALISED"; break;
    case JDMSOCKET_ERROR_WSAENETDOWN:                 TempText="JDMSOCKET_ERROR_WSAENETDOWN"; break;
    case JDMSOCKET_ERROR_WSAUNKNOWN:                  TempText="JDMSOCKET_ERROR_WSAUNKNOWN"; break;
    case JDMSOCKET_ERROR_GETHOSTBYNAME:               TempText="JDMSOCKET_ERROR_GETHOSTBYNAME"; break;
    case JDMSOCKET_ERROR_COULD_NOT_CREATE:            TempText="JDMSOCKET_ERROR_COULD_NOT_CREATE"; break;
    case JDMSOCKET_ERROR_COULD_NOT_BIND:              TempText="JDMSOCKET_ERROR_COULD_NOT_BIND"; break;
    case JDMSOCKET_ERROR_COULD_NOT_CONTROL_IOMODE:    TempText="JDMSOCKET_ERROR_COULD_NOT_CONTROL_IOMODE"; break;
    case JDMSOCKET_ERROR_COULD_NOT_LISTEN:            TempText="JDMSOCKET_ERROR_COULD_NOT_LISTEN"; break;
    case JDMSOCKET_ERROR_COULD_NOT_CLOSE:             TempText="JDMSOCKET_ERROR_COULD_NOT_CLOSE"; break;
    case JDMSOCKET_ERROR_COULD_NOT_CONNECT:           TempText="JDMSOCKET_ERROR_COULD_NOT_CONNECT"; break;
    case JDMSOCKET_ERROR_DURING_SEND:                 TempText="JDMSOCKET_ERROR_DURING_SEND"; break;
    case JDMSOCKET_ERROR_COULD_NOT_SELECT:            TempText="JDMSOCKET_ERROR_COULD_NOT_SELECT"; break;
    case JDMSOCKET_ERROR_COULD_NOT_ACCEPT:            TempText="JDMSOCKET_ERROR_COULD_NOT_ACCEPT"; break;
    case JDMSOCKET_ERROR_DURING_READ:                 TempText="JDMSOCKET_ERROR_DURING_READ"; break;
    case JDMSOCKET_ERROR_CONNECTION_CLOSED:           TempText="JDMSOCKET_ERROR_CONNECTION_CLOSED"; break;
    case JDMSOCKET_ERROR_LENGTH_DATA_EXCEEDED:        TempText="JDMSOCKET_ERROR_LENGTH_DATA_EXCEEDED"; break;
    case JDMSOCKET_ERROR_NO_FILEDESCRIPTOR:           TempText="JDMSOCKET_ERROR_NO_FILEDESCRIPTOR"; break;
    case JDMSOCKET_ERROR_NO_DATA:                     TempText="JDMSOCKET_ERROR_NO_DATA"; break;
    case JDMSOCKET_ERROR_USE_SOCKET_DEINIT:           TempText="JDMSOCKET_ERROR_USE_SOCKET_DEINIT"; break;
    case JDMSOCKET_ERROR_NO_DATA_TO_WRITE:            TempText="JDMSOCKET_ERROR_NO_DATA_TO_WRITE"; break;
    case JDMSOCKET_ERROR_NO_DATA_TO_READ:             TempText="JDMSOCKET_ERROR_NO_DATA_TO_READ"; break;
    case JDMSOCKET_ERROR_COULD_NOT_SET_SOCKET_OPTION: TempText="JDMSOCKET_ERROR_COULD_NOT_SET_SOCKET_OPTION"; break;
    default:                                          TempText="JDMSOCKET_UNKNOWN_ERROR"; break;
  }
  sprintf(FormatString, "(%+04.3ld) %.100s", ErrorValue, TempText.c_str());
  ErrorText=FormatString;

  return ErrorText.c_str();
} /* JDMSocket::GetErrorText */

/*---------------------------------------------------------------------------*/
/* JDMSocket::Listen                                                         */
/*---------------------------------------------------------------------------*/
long JDMSocket::Listen()
{
  WSASetLastError(0);
  SaveStatus(0);

  ServerFD=socket(AF_INET, SOCK_STREAM, 0);
  if(ServerFD==INVALID_SOCKET||ServerFD<=0) {
    ServerFD=0;
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_COULD_NOT_CREATE;
  }
  unsigned long OptionValue=1;
  if(setsockopt(ServerFD, SOL_SOCKET, SO_REUSEADDR, (char*)&OptionValue, sizeof(OptionValue))==SOCKET_ERROR) {
    SaveStatus(WSAGetLastError());
    DeInit();
    return JDMSOCKET_ERROR_COULD_NOT_SET_SOCKET_OPTION;
  }
  if(bind(ServerFD, (struct sockaddr*)&ServerAddress, sizeof(ServerAddress))==SOCKET_ERROR) {
    SaveStatus(WSAGetLastError());
    DeInit();
    return JDMSOCKET_ERROR_COULD_NOT_BIND;
  }
  if(SetBlockingMode()==SOCKET_ERROR) {
    SaveStatus(WSAGetLastError());
    DeInit();
    return JDMSOCKET_ERROR_COULD_NOT_CONTROL_IOMODE;
  }
  if(listen(ServerFD, 10)==SOCKET_ERROR) {
    SaveStatus(WSAGetLastError());
    DeInit();
    return JDMSOCKET_ERROR_COULD_NOT_LISTEN;
  }

  SaveStatus(WSAGetLastError());
  return JDMSOCKET_NO_ERROR;
} /* JDMSocket::Listen */

/*---------------------------------------------------------------------------*/
/* JDMSocket::Connect                                                        */
/*---------------------------------------------------------------------------*/
long JDMSocket::Connect()
{
  WSASetLastError(0);
  SaveStatus(0);

  ServerFD=socket(AF_INET, SOCK_STREAM, 0);
  if(ServerFD==INVALID_SOCKET||ServerFD<=0) {
    ServerFD=0;
    SaveStatus(WSAGetLastError());
    return JDMSOCKET_ERROR_COULD_NOT_CREATE;
  }
  if(SetBlockingMode()==SOCKET_ERROR) {
    SaveStatus(WSAGetLastError());
    DeInit();
    return JDMSOCKET_ERROR_COULD_NOT_CONTROL_IOMODE;
  }
  if(connect(ServerFD, (struct sockaddr*)&ServerAddress, sizeof(ServerAddress))==SOCKET_ERROR) {
    SaveStatus(WSAGetLastError());
    if(LastStatus!=WSAEWOULDBLOCK) { /* WSAEWOULDBLOCK could happen on a non-blocking socket */
      DeInit();
      return JDMSOCKET_ERROR_COULD_NOT_CONNECT;
    }

    WSASetLastError(0);
    fd_set FDS;
    long RetVal;
    struct timeval TimeOut;

    TimeOut.tv_sec = 1;
    TimeOut.tv_usec = 0;
    FD_ZERO(&FDS);
    FD_SET(ServerFD, &FDS);
    RetVal=select(sizeof(FDS)*8, 0, &FDS, 0, &TimeOut);
    if(RetVal==SOCKET_ERROR) {
      SaveStatus(WSAGetLastError());
      DeInit();
      return JDMSOCKET_ERROR_COULD_NOT_SELECT;
    }
    if(RetVal>0 && FD_ISSET(ServerFD, &FDS)) {
      /* Our socket is ready */
    }
    else {
      DeInit();
      return JDMSOCKET_ERROR_COULD_NOT_CONNECT;
    }
  }

  SaveStatus(WSAGetLastError());
  return JDMSOCKET_NO_ERROR;
} /* JDMSocket::Connect */

/*---------------------------------------------------------------------------*/
/* JDMSocket::SetBlockingMode                                                */
/*---------------------------------------------------------------------------*/
long JDMSocket::SetBlockingMode()
{
  /* SaveStatus should be done after calling this function */
  return ioctlsocket(ServerFD, FIONBIO, &BlockingMode);
} /* JDMSocket::SetBlockingMode */

/*---------------------------------------------------------------------------*/
/* JDMSocket::SaveStatus                                                     */
/*---------------------------------------------------------------------------*/
long JDMSocket::SaveStatus(long Value) {
  LastStatus=Value;
  return 0;
} /* JDMSocket::SaveStatus */

/*---------------------------------------------------------------------------*/
/* JDMSocket::InitSocketLib                                                  */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::InitSocketLib()
{
  WORD    RequestedVersion = MAKEWORD(1,1);
  WSADATA wsaData;

  switch(WSAStartup(RequestedVersion, &wsaData)) {
    case 0:                  return JDMSOCKET_NO_ERROR; break;
    case WSASYSNOTREADY:     return JDMSOCKET_ERROR_WSASYSNOTREADY; break;
    case WSAVERNOTSUPPORTED: return JDMSOCKET_ERROR_WSAVERNOTSUPPORTED; break;
    case WSAEINPROGRESS:     return JDMSOCKET_ERROR_WSAEINPROGRESS; break;
    case WSAEPROCLIM:        return JDMSOCKET_ERROR_WSAEPROCLIM; break;
    case WSAEFAULT:          return JDMSOCKET_ERROR_WSAEFAULT; break;
    default:                 return JDMSOCKET_ERROR_WSAUNKNOWN; break;
  }
} /* JDMSocket::InitSocketLib */

/*---------------------------------------------------------------------------*/
/* JDMSocket::DeInitSocketLib                                                */
/*---------------------------------------------------------------------------*/
long JDMSOCKET_CALL_CONV JDMSocket::DeInitSocketLib()
{
  switch(WSACleanup()) {
    case 0:                  return JDMSOCKET_NO_ERROR; break;
    case WSANOTINITIALISED:  return JDMSOCKET_ERROR_WSANOTINITIALISED; break;
    case WSAENETDOWN:        return JDMSOCKET_ERROR_WSAENETDOWN; break;
    case WSAEINPROGRESS:     return JDMSOCKET_ERROR_WSAEINPROGRESS; break;
    default:                 return JDMSOCKET_ERROR_WSAUNKNOWN; break;
  }
} /* JDMSocket::DeInitSocketLib */
